package convertquery

import (
	"encoding/json"
	"fmt"
)

// Service is a model for the convertquery use case
type Service struct {
}

// NewService creates a new convertquery service
func NewService() *Service {
	return &Service{}
}

// Constraint datatypes
// text		  MatchTypes: exact/contains/startswith/endswith
// number   MatchTypes: GT/LT/EQ
// bool     MatchTypes: EQ/NEQ

// Ranges dus tussen 10 half 5 bijv.

// Struct used for JSON conversion
type parsedJSON struct {
	Return    returnStruct     //`json:"return"`
	Entities  []entityStruct   //`json:"entities"`
	Relations []relationStruct //`json:"relations"`
}

type returnStruct struct {
	Entities  []int
	Relations []int
}

type entityStruct struct {
	Type        string
	Constraints []constraintStruct
}
type relationStruct struct {
	Type        string
	EntityFrom  int
	EntityTo    int
	Depth       searchDepthStruct
	Constraints []constraintStruct
}
type searchDepthStruct struct {
	Min int
	Max int
}

type constraintStruct struct {
	Attribute string
	Value     string
	DataType  string
	MatchType string
}

// ConvertQuery converts a json string to an AQL query
func (s *Service) ConvertQuery(jsonMsg *[]byte) (*string, error) {

	jsonStruct, err := convertJSONToStruct(jsonMsg)
	if err != nil {
		fmt.Println(err)
		return nil, err
	}

	//Per node query
	//	per constraint
	//relations koppelen ze samen
	//return statement

	// var result *string
	// if len(jsonStruct.Return.Relations) > 0 {
	// 	result = createEdgeQuery(jsonStruct, "e0")
	// } else {

	// 	result = createAllNodesQuery(jsonStruct.Return.Entities, jsonStruct.Entities)
	// }
	result := createQuery(jsonStruct)
	return result, nil
}

func convertJSONToStruct(jsonMsg *[]byte) (*parsedJSON, error) {
	jsonStruct := parsedJSON{}
	err := json.Unmarshal(*jsonMsg, &jsonStruct)

	if err != nil {
		return nil, err
	}

	return &jsonStruct, nil
}

func createQuery(jsQuery *parsedJSON) *string {
	// GROTE SIDENOTE:
	// Vrij zeker dat een query waar alléén edges worden opgevraagd (#4)
	// niet wordt gesupport door zowel de result parser als de frontend reciever

	jsonQuery := *jsQuery

	// TODO:
	// NODES
	// EDGES (als ze er zijn)
	// RETURN STATEMENT
	var ret string
	for k, v := range jsonQuery.Entities {
		name := fmt.Sprintf("n%v", k)
		ret += *createNodeLet(&v, &name)
	}

	var onlyEdge bool
	if len(jsonQuery.Entities) == 0 {
		onlyEdge = true
	}

	for k, v := range jsonQuery.Relations {
		name := fmt.Sprintf("e%v", k)
		ret += *createEdgeLet(&v, &name, onlyEdge)
	}

	finalReturn := "RETURN {"

	// If there is an edge you can only return the edge, because the nodes are included.
	// If there are no edges, return the nodes
	if l := len(jsonQuery.Relations); l > 0 {
		for k := range jsonQuery.Relations {
			finalReturn += fmt.Sprintf("e%v:e%v", k, k)
			if k < l-1 {
				finalReturn += ","
			}
		}
	} else {
		for k := range jsonQuery.Entities {
			finalReturn += fmt.Sprintf("n%v:n%v", k, k)
			if k < l-1 {
				finalReturn += ","
			}
		}
	}

	finalReturn += "}"
	ret += finalReturn
	return &ret
}

func createNodeLet(node *entityStruct, name *string) *string {
	header := fmt.Sprintf("LET %v = (\n\tFOR x IN %v \n", *name, node.Type)
	footer := "\tRETURN x\n)\n"

	constraints := *createConstraints(&node.Constraints, false, false)

	ret := header + constraints + footer
	return &ret
}

func createEdgeLet(edge *relationStruct, name *string, onlyEdge bool) *string {
	var (
		header        string
		forEdge       string
		forSecondNode string
	)

	footer := "\n\tLIMIT 100\n\tRETURN { vertices: p.vertices[*], edges: p.edges[*] }\n)\n"

	// WICKED SWITCHES LETSAGO
	if edge.EntityFrom != -1 {
		// # 1 (2)
		header = fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", *name, edge.EntityFrom)
		forEdge = fmt.Sprintf("\tFOR v, e, p IN %v..%v OUTBOUND x %s \n", edge.Depth.Min, edge.Depth.Max, edge.Type)
		if edge.EntityTo != -1 {
			// # 2
			forSecondNode = fmt.Sprintf("\tFILTER v IN n%v \n", edge.EntityTo)
		}
	} else {
		if edge.EntityTo != -1 {
			// # 3
			header = fmt.Sprintf("LET %v = (\n\tFOR x IN n%v \n", *name, edge.EntityTo)
			forEdge = fmt.Sprintf("\tFOR v, e, p IN %v..%v INBOUND x %s \n", edge.Depth.Min, edge.Depth.Max, edge.Type)
		} else {
			// # 4
			header = fmt.Sprintf("LET %v = (\n\tFOR x IN %v \n", *name, edge.Type)
			footer = "\tLIMIT 100\n\tRETURN x\n)\n"
		}
	}

	constraints := *createConstraints(&edge.Constraints, true, onlyEdge)

	ret := header + forEdge + forSecondNode + constraints + footer
	return &ret
}

/*
#1
{
  "Return": {
    "Entities": [
      0,
      1
    ],
    "Relations": [
      0
    ]
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": []
    },
    {
      "Type": "airports",
      "Constraints": []
    }
  ],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": 0,
      "EntityTo": 1,
      "Constraints": []
    }
  ]
}
*/

/*
#2

{
  "Return": {
    "Entities": [
      0
    ],
    "Relations": [
      0
    ]
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": []
    }
  ],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": 0,
      "EntityTo": -1,
      "Constraints": []
    }
  ]
}
*/

/*
#3
{
  "Return": {
    "Entities": [
      0
    ],
    "Relations": [
      0
    ]
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": []
    }
  ],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": -1,
      "EntityTo": 0,
      "Constraints": []
    }
  ]
}
*/

/*
#4
{
  "Return": {
    "Entities": [],
    "Relations": [
      0
    ]
  },
  "Entities": [],
  "Relations": [
    {
      "Type": "flights",
      "Depth": {
        "min": 1,
        "max": 1
      },
      "EntityFrom": -1,
      "EntityTo": -1,
      "Constraints": [
        {
          "Attribute": "Month",
          "Value": "1",
          "DataType": "number",
          "MatchType": "EQ"
        },
        {
          "Attribute": "Day",
          "Value": "15",
          "DataType": "number",
          "MatchType": "EQ"
        }
      ]
    }
  ]
}
*/

/*
#5
{
  "Return": {
    "Entities": [
      0
    ],
    "Relations": []
  },
  "Entities": [
    {
      "Type": "airports",
      "Constraints": [
        {
          "Attribute": "city",
          "Value": "New York",
          "DataType": "text",
          "MatchType": "exact"
        },
        {
          "Attribute": "country",
          "Value": "USA",
          "DataType": "text",
          "MatchType": "exact"
        }
      ]
    }
  ],
  "Relations": []
}
*/
